unit VuMeter;

Interface

Uses
  WinDows, Messages, SysUtils, Classes, Graphics, Controls, Math, MMSystem;

//****************************************************************************************************

Const
  MaxBuffers = 4;

Type
  TBufferEvent = Procedure(Buffer: pChar; Var Size: Integer; Sender: TObject) Of Object;
  TVuInputEvent = Procedure(ChLeft, ChRight: Real; Sender: TObject) Of Object;

Type
  TVuMeter = Class(TCusTomControl)
  Private
    { Private Declarations }
    FBufferSize: Integer;
    FRequestedBufferSize: Integer;
    FNumBuffers: Integer;
    FPaused: Boolean;

    WaveDeviceOpen: Boolean;
    FOnBufferFilled: TBufferEvent;
    FVuInput: TVuInputEvent;

    FWaveFmtEx: TWaveFormatEx;
    FWaveDevice: DWOrd;
    hWaveHeader	: Array[0..MaxBuffers - 1] Of THandle;
    WaveHdr: Array[0..MaxBuffers - 1] Of PWaveHdr;
    WaveBuffer: Array[0..MaxBuffers - 1] Of Lpstr;
    hWaveBuffer: Array[0..MaxBuffers - 1] Of THandle;
    BufIndex: Integer;
    ContinueProcessing: Boolean; 

    FBackGRoundColor: tColor;
    FLeft: single;
    FRight: single;
    FLeftIndex: Integer;
    FRightIndex: Integer;

    FMax: Integer;
    FMin: Integer;
    FHighLevel: Integer;
    FMidLevel: Integer;
    FPenWidth : Integer;

    FStereo: Boolean;
    Ch1: Array[0..1] Of Integer;
    Ch2: Array[0..1] Of Integer;
    DelayLine: Array[0..1] Of Real;
    Current: Array[0..1] Of Integer;

    FHigh, FDHigh, FMid, FDMid, FLow, FDLow: TColor;
    Levels: Array[0..100] Of single;
    FHighColor: Array[0..100] Of TColor;
    FDimColor: Array[0..100] Of TColor;
    NumberLevels: Integer;
    Procedure MakeWaveFmtConsistent;

    Function QueueBuffer: Boolean;
    Function ProcessBuffer(B: Lpstr; N: Integer): Boolean;
    Procedure CloseWaveDevice;
    Function Start: Boolean;
    Function Setup: Boolean;
    Procedure SetBackgRoundColor(Value: tColor);
    Procedure ReSize;
    Procedure SetLevels;
    Procedure DrawHighLine(Line: Integer; Dimmed: Boolean; Channel: Integer);
    Procedure DrawDelayLine(Line: Integer; Channel: Integer);
    Procedure GetChannel(Value: Single; Channel: Integer);
  Protected
    { Protected Declarations }
    DrawWidth: Integer;
    DrawHeight: Integer;
    Bitmap: TBitmap;
    ErrorMessage: String;
    Active: Boolean;
    FilledBuffers : Integer;
    QueuedBuffers : Integer;
    ProcessedBuffers: Integer;
    WaveHandle: HWaveOut;

    Function InitWaveHeaders: Boolean;
    Function AllocWaveBuffers: Boolean;
    Function FreeWaveBuffers: Boolean;
    Function AllocWaveHeaders: Boolean;
    Procedure FreeWaveHeaders;
 
    Procedure Paint; Override;
    Procedure WMEraseBkgnd(Var M: TWMEraseBkgnd); Message WM_ERASEBKGND;
    Procedure ReDraw;

    Procedure SetMax(Max: Integer);
    Procedure SetMin(Min: Integer);

    Procedure SetHighLevel(Level: Integer);
    Procedure SetMidLevel(Level: Integer);

    Procedure SetHighColor(Col: TColor);
    Procedure SetMidColor(Col: TColor);
    Procedure SetLowColor(Col: TColor);
    Procedure SetColors;
    Procedure SetStereo(Value: Boolean);
    Procedure SetPenWidth(Value: Integer);
    Procedure StopAtOnce; Virtual;
    Procedure StopGracefully; Virtual;
    Procedure ClearBmp;
    Procedure SetVuMeter(Buffer: PChar; Size: Integer);
  Public
    { Public Declarations }
    ConstrucTor Create(aOwner: TComponent); Override;
    DestrucTor Destroy; Override;
  Published
    { Published Declarations }
    Property MaxLevel: Integer Read FMax Write SetMax;
    Property LowLevel: Integer Read FMin Write SetMin;

    Property HighLevel: Integer Read FHighLevel Write SetHighLevel;
    Property MidLevel: Integer Read FMidLevel Write SetMidLevel;

    Property HighColor: TColor Read FHigh Write SetHighColor;
    Property MidColor: TColor Read FMid Write SetMidColor;
    property LowColor: TColor Read FLow Write SetLowColor;

    Property Stereo: Boolean Read FStereo Write SetStereo;
    Property BackGRoundColor: tColor Read FBackGRoundColor Write SetBackgRoundColor;

    Property OnBufferFilled: TBufferEvent Read FOnBufferFilled Write FOnBufferFilled;
    Property OnVuInput: TVuInputEvent Read FVuInput Write FVuInput;
  end;

//****************************************************************************************************
  TWaveIn = Class(TWinControl)
  Private
    { Private Declarations }
    AudioComponent: TVuMeter;
    Procedure BufferFinished(Var Msg: TMessage); Message MM_WIM_DATA;
  end;
//****************************************************************************************************

Procedure Register;

Var
WaveIn: TWaveIn;

Implementation

Function TWaveInGetErrorText(iErr: Integer): String;
Var
  ErrorMsg: Array[0..255] Of Char;
Begin
  WaveInGetErrorText(iErr, ErrorMsg, SizeOf(ErrorMsg));
  Result := StrPas(ErrorMsg);
end;

Procedure TVuMeter.MakeWaveFmtConsistent;
Begin
  With FWaveFmtEx Do
  Begin
    nBlockAlign := (wBitsPerSample Div 8) * nChannels;
    nAvgBytesPerSec := nSamplesPerSec * nBlockAlign;
  end;
  FBufferSize := FRequestedBufferSize - (FRequestedBufferSize Mod FWaveFmtEx.nBlockAlign);
end;

Function TVuMeter.InitWaveHeaders: Boolean;
Var
  I: Integer;
Begin
  MakeWaveFmtConsistent;
  For I := 0 To FNumBuffers - 1 Do
    With WaveHdr[I]^ Do
    Begin
      lpData := WaveBuffer[I];
      dwBufferLength := FBufferSize;
      dwBytesRecorded := 0;
      dwUser := 0;
      dwFlags := 0;
      dwLoops := 0;
      lpNext := Nil;
      Reserved := 0;
    end;
  InitWaveHeaders := True;
end;

Function TVuMeter.AllocWaveBuffers: Boolean;
Var
  I: Integer;
Begin
  For I := 0 To fNumBuffers - 1 Do
  Begin
    hWaveBuffer[I] := GlobalAlloc(GMEM_MOVEABLE Or GMEM_SHARE, fBufferSize);
    If (hWaveBuffer[I] = 0) Then
    Begin
      FreeWaveBuffers;
      AllocWaveBuffers := False;
      Exit;
    end;
    WaveBuffer[I] := GlobalLock(hWaveBuffer[I]);
    If (WaveBuffer[I] = Nil) Then
    Begin
      FreeWaveBuffers;
      AllocWaveBuffers := False;
      Exit;
    end;
    WaveHdr[I].lpData := WaveBuffer[I];
  end;
  AllocWaveBuffers := True;
end;

Function TVuMeter.FreeWaveBuffers: Boolean;
Var
  I: Integer;
Begin
  Result := False;
  For I := 0 To MaxBuffers - 1 Do
  Begin
    If (hWaveBuffer[I] <> 0) Then
    Begin
      GlobalUnlock(hWaveBuffer[I]);
      GlobalFree(hWaveBuffer[I]);
      hWaveBuffer[I] := 0;
      WaveBuffer[I] := Nil;
      Result := True;
    end;
  end;
end;

Function TVuMeter.AllocWaveHeaders: Boolean;
Var
  I: Integer;
Begin
  For I := 0 To MaxBuffers - 1 Do
  Begin
    hWaveHeader[I] := GlobalAlloc(GMEM_MOVEABLE Or GMEM_SHARE Or GMEM_ZEROINIT, SizeOf(TWAVEHDR));
    If (hWaveHeader[I] = 0) Then
    Begin
      FreeWaveHeaders;
      ErrorMessage := 'Error allocating wave header memory';
      AllocWaveHeaders := False;
      Exit;
    end;

    WaveHdr[I] := GlobalLock(hWaveHeader[I]);
    If (WaveHdr[I] = Nil) Then
    Begin
      FreeWaveHeaders;
      AllocWaveHeaders := False;
      Exit;
    end;
  end;
  AllocWaveHeaders := True;
end;

Procedure TVuMeter.FreeWaveHeaders;
Var
  I: Integer;
Begin
  For I := 0 To MaxBuffers - 1 Do
  Begin
    If (hWaveHeader[I] <> 0) Then
    Begin
      GlobalUnlock(hWaveHeader[I]);
      GlobalFree(hWaveHeader[I]);
      hWaveHeader[I] := 0;
      WaveHdr[I] := Nil;
    end
  end;
end;

Procedure TVuMeter.CloseWaveDevice;
Var
  I: Integer;
Begin
  Active := False;
  For I := 0 To FNumBuffers - 1 Do
  WaveInUnprepareHeader(WaveHandle, WaveHdr[I], SizeOf(TWAVEHDR));
  WaveInReSet(WaveHandle);
  WaveInClose(WaveHandle);
  WaveDeviceOpen := False;
end;

Function TVuMeter.Setup: Boolean;
Var
  iErr: Integer;
  I: Integer;
Begin
  If (WaveDeviceOpen) Then
  Begin
    Result := False;
    Exit;
  end;
  BufIndex := 0;
  WaveIn := TWaveIn.CreateParented(TWinControl(Owner).Handle);
  WaveIn.Visible := False;
  WaveIn.AudioComponent := Self;

  iErr := WaveInOpen(@WaveHandle, FWaveDevice, @FWaveFmtEx, Integer(WaveIn.Handle),
    0, CALLBACK_WINDOW Or WAVE_ALLOWSYNC);

  If (iErr <> 0) Then
  Begin
    Result := False;
    Exit;
  end;
  WaveDeviceOpen := True;
  If (Not InitWaveHeaders) Then
  Begin
    Result := False;
    Exit;
  end;
  For I := 0 To FNumBuffers - 1 Do
  Begin
    iErr := WaveInPrepareHeader(WaveHandle, WaveHdr[I], SizeOf(TWAVEHDR));
    If (iErr <> 0) Then
    Begin
      CloseWaveDevice;
      Result := False;
      Exit;
    end;
  end;
  QueuedBuffers := 0;
  ProcessedBuffers := 0;
  FilledBuffers := 0;
  ContinueProcessing := True;
  Active := True;
  Result := True;
end;

Function TVuMeter.QueueBuffer: Boolean;
Var
  iErr: Integer;
Begin
  WaveHdr[BufIndex].dwFlags := WHDR_PREPARED;
  iErr := WaveInAddBuffer(WaveHandle, WaveHdr[BufIndex], SizeOf(TWAVEHDR));
  If (iErr <> 0) Then
  Begin
    StopGracefully;
    Result := False;
    Exit;
  end;
  BufIndex := (BufIndex + 1) Mod FNumBuffers;
  QueuedBuffers := QueuedBuffers + 1;
  Result := True;
end;

Function TVuMeter.Start: Boolean;
Var
  I, iErr: Integer;
Begin
  iErr := WaveInStart(WaveHandle);
  If (iErr <> 0) Then
  Begin
    CloseWaveDevice;
    Result := False;
    Exit;
  end;
  Active := True;
  For I := 0 To FNumBuffers - 1 Do
    If (Not QueueBuffer) Then
    Begin
      CloseWaveDevice;
      Result := False;
      Exit;
    end;
  Result := True;
end;

Function TVuMeter.ProcessBuffer(B: Lpstr; N: Integer): Boolean;
Begin
Result := ContinueProcessing;
  If (Result) Then
  Begin
  SetVuMeter(B, N);
  If Assigned(FOnBufferFilled) Then
   FOnBufferFilled(B, N, Nil);
  end;
end;

Procedure TVuMeter.StopAtOnce;
Begin
  If (Not WaveDeviceOpen) Then Exit;
  Active := False;
  ContinueProcessing := False;
  WaveInReSet(WaveHandle);
  CloseWaveDevice;
end;

Procedure TVuMeter.StopGracefully;
Begin
  If (Not WaveDeviceOpen) Then Exit;
  ContinueProcessing := False;
end;

Procedure TWaveIn.BufferFinished(Var Msg: TMessage);
Begin
  With AudioComponent Do
  Begin
    ProcessedBuffers := ProcessedBuffers + 1;
    QueuedBuffers := QueuedBuffers - 1;
    Active := (QueuedBuffers > 0);
    If (ProcessBuffer(WaveBuffer[BufIndex], FBufferSize)) Then QueueBuffer;
    If (Not Active) Then
    Begin
      ContinueProcessing := False;
      CloseWaveDevice;
    end;
  end;
end;

DestrucTor TVuMeter.Destroy;
Begin
  Bitmap.Free;
  ContinueProcessing := False;
  CloseWaveDevice;
  FreeWaveBuffers;
  FreeWaveHeaders;
  Inherited Destroy;
end;

Procedure TVuMeter.ReSize;
Begin
  If Not (Bitmap = Nil) Then
    If Not ((DrawWidth = Width) And (DrawHeight = Height)) Then
    Begin
      Bitmap.Width := Width;
      Bitmap.Height := Height;
      DrawWidth := Width;
      DrawHeight := Height;
      SetLevels;
      ReDraw;
      InValidate;
    end;
end;

Procedure TVuMeter.ClearBmp;
Var NewRect: TRect;
Begin
  Bitmap.Canvas.Brush.Color := FBackGRoundColor;
  Bitmap.Canvas.Brush.Style := bsSolid;
  NewRect := Rect(0, 0, DrawWidth, DrawHeight);
  Bitmap.Canvas.FillRect(NewRect);
end;

Procedure TVuMeter.Paint;
Begin
  If Not ((DrawWidth = Width) And (DrawHeight = Height)) Then
    ReSize
  Else
  Begin
    Canvas.CopyMode := cmSrcCopy;
    Canvas.Draw(0, 0, Bitmap);
  end;
end;

Procedure TVuMeter.SetBackgRoundColor(Value: tColor);
Begin
  FBackGRoundColor := Value;
  ReDraw;
  inValidate;
end;

Procedure TVuMeter.WMEraseBkgnd(Var M: TWMEraseBkgnd);
Begin
  M.Result := LResult(False);
end;

Function SetShaDow(Color: TColor): TColor;
Type
  ColorConvert = Record
    Case Byte Of
      0: (Z: TColor);
      1: (A, B, C, D: Byte);
  end;
Begin
  ColorConvert(Result).A := ColorConvert(Color).A Div 2;
  ColorConvert(Result).B := ColorConvert(Color).B Div 2;
  ColorConvert(Result).C := ColorConvert(Color).C Div 2;
  ColorConvert(Result).D := ColorConvert(Color).D Div 2;
end;

Procedure TVuMeter.SetMax(Max: Integer);
Begin
  FMax := Max;
  If (FMax < FMin) Then
    FMax := FMin + 1;
  SetLevels;
end;

Procedure TVuMeter.SetMin(Min: Integer);
Begin
  FMin := Min;
  If FMin > FMax Then
    FMin := FMax - 1;
  SetLevels;
end;

Procedure TVuMeter.SetHighLevel(Level: Integer);
Begin
  FHighLevel := Level;
  If FHighLevel > FMax Then
    FHighLevel := FMax;
  If FHighLevel < FMin Then
    FHighLevel := FMin;
  SetLevels;
end;

Procedure TVuMeter.SetMidLevel(Level: Integer);
Begin
  FMidLevel := Level;
  If FMidLevel > FMax Then
    FMidLevel := FMax;
  If FMidLevel < FMin Then
    FMidLevel := FMin;
  SetLevels;
end;

Procedure TVuMeter.SetLevels;
Var
  I: Integer;
  Range: Integer;
  LevelMid: Integer;
  LevelHigh: Integer;
  XMiddle : Integer;
Begin
  NumberLevels := ((Height - 2) Div 3) - 1;

  If NumberLevels = 0 Then
    NumberLevels := 1;
  If NumberLevels > 100 Then
    NumberLevels := 100;

  Range := FMax - FMin;
  If Range < 1 Then Range := 1;

  LevelMid := NumberLevels - Round((FMax - FMidLevel) * NumberLevels / Range);
  LevelHigh := NumberLevels - Round((FMax - FHIghLevel) * NumberLevels / Range);
Inc(NumberLevels);
  For I := NumberLevels DownTo 0 Do
  Begin
    Levels[I] := Power(10, -1.0 * (NumberLevels - I) / NumberLevels * (Range / 20.0));
    If I > LevelHigh Then
    Begin
      FHighColor[I] := FHigh;
      FDimColor[I] := FDHigh;
    end
    Else If I > LevelMid Then
    Begin
      FHighColor[I] := FMid;
      FDimColor[I] := FDMid;
    end
    Else
    Begin
      FHighColor[I] := FLow;
      FDimColor[I] := FDLow;
    end;
  end;

  If FStereo Then
    XMiddle := Width Div 2
  Else
    XMiddle := Width;
  Ch1[0] := 1;
  Ch2[0] := XMiddle - 2;
  Ch1[1] := XMiddle + 1;
  Ch2[1] := Width - 2 ;
  ReDraw;
end;

Procedure TVuMeter.SetHighColor(Col: TColor);
Begin
  FHigh := Col;
  SetColors;
end;

Procedure TVuMeter.SetMidColor(Col: TColor);
Begin
  FMid := Col;
  SetColors;
end;

Procedure TVuMeter.SetLowColor(Col: TColor);
Begin
  FLow := Col;
  SetColors;
end;

Procedure TVuMeter.SetColors;
Begin
  FDHigh := SetShaDow(FHigh);
  FDMid := SetShaDow(FMid);
  FDLow := SetShaDow(FLow);
  ReDraw;
end;

Procedure TVuMeter.SetStereo(Value: Boolean);
Begin
  If Not (Value = FStereo) Then
  Begin
    FStereo := Value;
    SetLevels;
  end;
  If Value Then
    FWaveFmtEx.nChannels := 2
  Else
    FWaveFmtEx.nChannels := 1;
  MakeWaveFmtConsistent;
end;

Procedure TVuMeter.SetPenWidth(Value: Integer);
Begin
  If Not (Value = FPenWidth) Then
  Begin
    FPenWidth := Value;
    ReDraw;
  end;
end;

ConstrucTOr TVuMeter.Create(aOwner: TComponent);
Var
  I: Integer;
Begin
  Inherited Create(AOwner);
  FNumBuffers := 2;
  FRequestedBufferSize := 8192;
  Active := False;
  FPaused := False;
  FWaveDevice := WAVE_MAPPER;
  FWaveFmtEx.wFormatTag := WAVE_FORMAT_PCM;
  FWaveFmtEx.wBitsPerSample := 16;
  FWaveFmtEx.nChannels := 2;
  FWaveFmtEx.nSamplesPerSec := 48000;
  MakeWaveFmtConsistent;
  For I := 0 To MaxBuffers - 1 Do WaveBuffer[I] := Nil;
  AllocWaveHeaders;
  AllocWaveBuffers;

  Bitmap := TBitmap.Create;
  Width := 16;
  Height := 98;
  BackGRoundColor := clBlack;
  ReSize;
  ClearBmp;
  DrawWidth := Width;
  DrawHeight := Height;
  FLeft := 0.0;
  FRight := 0.0;
  FLeftIndex := 0;
  FRightIndex := 0;
  FPenWidth := 2;
  FHigh := clRed;
  FMid := clYellow;
  FLow := clLime;
  SetColors;
  FMax := 0;
  FMin := -40;
  FHighLevel := -3;
  FMidLevel := -12;
  FStereo := False;
  SetStereo(True);
If Setup Then Start;
GetChannel(0.9, 0);
GetChannel(0.9, 1);
end;

Procedure TVuMeter.DrawHighLine(Line: Integer; Dimmed: Boolean; Channel: Integer);
Begin
  With Bitmap.Canvas Do
  Begin
    Pen.Width := 2;
   If Dimmed Then
      Pen.Color := FDimColor[Line]
    Else
      Pen.Color := FHighColor[Line];
    MoveTo(Ch1[Channel], Height - (Line * 3) - 2);
    LineTo(Ch2[Channel], Height - (Line * 3) - 2);
  end;
end;

Procedure TVuMeter.DrawDelayLine(Line: Integer; Channel: Integer);
Begin
  With Bitmap.Canvas Do
  Begin
    Pen.Width := 2;
    Pen.Color := FHighColor[Line];
    MoveTo(Ch1[Channel], Height - (Line * 3) - 2);
    LineTo(Ch2[Channel], Height - (Line * 3) - 2);
    Inc(Line);
    Pen.Color := FDimColor[Line];
    MoveTo(Ch1[Channel], Height - (Line * 3) - 2);
    LineTo(Ch2[Channel], Height - (Line * 3) - 2);
  end;
end;

Procedure TVuMeter.ReDraw;
Var
  I: Integer;
Begin
  If Not ((Bitmap = Nil) Or (Parent = Nil)) Then
  Begin
    ClearBmp;
    For I := 0 To NumberLevels +2 Do
    Begin
      DrawHighLine(I +1, True, 0);
      If FStereo Then
        DrawHighLine(I +1, True, 1);
    end;
  end;
  Current[0] := 0;
  Current[1] := 0;
end;

Procedure TVuMeter.GetChannel(Value: Single; Channel: Integer);
Var
  I: Integer;
Begin
  I := Current[Channel];
  If Value > Levels[I] Then
  Begin
    While ((I < NumberLevels) And (Value > Levels[I])) Do
    Begin
      DrawHighLine(I, False, Channel);
    Inc(I);
      If DelayLine[Channel] < I Then
      DelayLine[Channel] := I;
    end;
  end
  Else
  Begin
    While ((I > 0) And (Value <= Levels[I])) Do
    Begin
      DrawHighLine(I, True, Channel);
      Dec(I);
    end;
  end;
  Current[Channel] := I;
end;

Procedure TVuMeter.SetVuMeter(Buffer: PChar; Size: Integer);
Type
 PSample = ^SmallInt;
Var
I : Integer;
P : PSample;
Begin
   P := PSample(buffer);
  FLeft := 0;
  FRight := 0;
  For I := 0 To (Size Div 4) - 1 Do
  Begin
  If P^ > FLeft Then
  FLeft := P^;
  Inc(P);
  If P^ > FRight Then
  FRight := P^;
  Inc(P);
  end;

GetChannel(0.78 * ((FLeft-800)  / 32767), 0);
If FStereo Then
GetChannel(0.78 * ((FRight-800)  / 32767), 1);
FLeft := (FLeft  / 32767);
FRight := (FRight  / 32767);
DrawDelayLine(Round(DelayLine[0]), 0);
DrawDelayLine(Round(DelayLine[1]), 1);
Paint;
If Assigned(FVuInput) Then
FVuInput(FLeft, FRight, Self);
If DelayLine[0] >-1 Then
DelayLine[0] := DelayLine[0] - 0.0625;
If DelayLine[1] >-1 Then
DelayLine[1] := DelayLine[1] - 0.0625;
end;

Procedure Register;
Begin
  RegisterComponents('GMS', [TVuMeter]);
end;

end.